// Author: Sam Cohen
// Date: 1/11/25
// File Name: serial.c
// Desc: Implementation of serial output from AVR microcontroller.

#ifndef SERIAL_C
#define SERIAL_C

#include "serial.h"
#include "avr_macros.h"


// Sources:
// 1. ATMEGA32p data sheet
// This blog helped a ton
// 2. /https://www.appelsiini.net/2011/simple-usart-with-avr-libc/ 
// This paper was incredibly helpful in understanding UART
// 3. https://www.analog.com/en/resources/analog-dialogue/articles/uart-a-hardware-communication-protocol.html 

size_t init_UART(unsigned long baud_rate) 
{

	// first we calculate the UBRR0 registers value using our desired baud rate
	// this represents an interal register that manages the timing of UART
	// The formula is [CLOCKSPEED/( 16 * desired baud)] - 1

	// alot of this baud rate calc should be moved into a macro.
	unsigned long raw_UBRR_val = ((F_CPU + 8UL * (baud_rate)) / (16UL * baud_rate)) - 1UL;

	// We need to test whether this apporximation of best best possible UBRR value is good
	// enough, within 2% for some reason, so desycronization does not become a significant problem
	// to do this we recalculate clock the clock speed using the derived ubrr and look at error
	// if there is error we turn on a special setting called 2x speed mode which changes the clock speed when
	// transmitting and allows for a chance at using the desired baud rate again.
	// my understanding of this comes from the set baud library and the atmega doc so I am probably missing some
	// crucial details.
	unsigned long highest_acceptable_clock_speed = (baud_rate + (0.02) * baud_rate) * (16 * (raw_UBRR_val + 1));
	unsigned long lowest_acceptable_clock_speed = (baud_rate - (0.02) * baud_rate) * (16 * (raw_UBRR_val + 1));

	if (F_CPU > highest_acceptable_clock_speed || F_CPU < lowest_acceptable_clock_speed) 
	{
		//((F_CPU + 8UL * (baud_rate)) / (16UL * baud_rate)) - 1UL;

		setRH(UCSR0A, U2X0); // change to 2x speed mode

		raw_UBRR_val = ((F_CPU + 4UL * (baud_rate)) / (8UL * baud_rate)) - 1UL;
		unsigned long highest_acceptable_clock_speed = (baud_rate + (0.02) * baud_rate) * (8 * (raw_UBRR_val + 1));
		unsigned long lowest_acceptable_clock_speed = (baud_rate - (0.02) * baud_rate) * (8 * (raw_UBRR_val + 1));
		
		if (F_CPU > highest_acceptable_clock_speed || F_CPU < lowest_acceptable_clock_speed)
			return 0;  // no acceptable UBRR value was found for this baud rate, we accept defeat and return false.
	} 
	else
		setRL(UCSR0A, U2X0); // change to 2x speed mode



	// We need to break this value down into its 12 LSB as that is what avr expects
	// except we don't know the endianess of our system so we have to do this:
	uint16_t value = 0x1234;
	uint8_t first_byte = ((uint8_t *)&value)[0];


	if (first_byte == 0x34)  // is little e
	{
		UBRR0L = ((char *)(&raw_UBRR_val))[0];         // sets lower eight bits of 12 bit baud rate
		UBRR0H = ((char *)(&raw_UBRR_val))[1] & 0x0f;  // sets upper four bits of 12 bit baud rate
	} 
	else 
	{
		// sets lower eight bits of 12 bit baud rate
		setR(UBRR0L,((char *)(&raw_UBRR_val))[sizeof(unsigned long) - 1]);         
		// sets upper four bits of 12 bit baud rate
		setR(UBRR0H, ((char *)(&raw_UBRR_val))[sizeof(unsigned long) - 2] & 0x0f);  
	}


	// Turn the reciever and transmitter on.
	setRH(UCSR0B, RXEN0);
	setRH(UCSR0B, TXEN0);

	// Sets the receiver and transmitter to 8 bit data setting.
	setRH(UCSR0C, UCSZ01);
	setRH(UCSR0C, UCSZ00);
	

	// Turns on parity checking so if something goes really wrong
	// when sending or recieving there is a flag that can be checked
	setRH(UCSR0C, UPM01);
	setRH(UCSR0C, UPM00);

	return 1;
}

int transmit_byte_UART(unsigned char data_byte) 
{
	//this line is required if we want to use it with the stdin header
	if (data_byte == '\n')
		transmit_byte_UART('\r');

	// UCSR0A register has a bunch of status info
	// the UDRE0 bit of the register is the USART Data Register Empty
	// bit, and will be high when the register is empty and ready to recieve new data to send.
	while (!(peekR(UCSR0A, UDRE0)))  //while not ready to transieve
	{
		//no op;
	}

	UDR0 = data_byte;
	if(0) //if we detect an error return false
		return EOF;
	return data_byte;
}

size_t UART_printf(const char* text, ...)
{
	va_list v_list_pointer;
	va_start(v_list_pointer, text);

	char string_buffer[128];

	size_t result = vsprintf(string_buffer, text, v_list_pointer);
	if(result >= 128)
		return 0;
	va_end(v_list_pointer);
	for(size_t i = 0; i < result; i++)
	{
		char transmission_success = transmit_byte_UART(string_buffer[i]);
	}
	return 1;

}


unsigned char receive_byte_UART()
{

	// UCSR0A register has a bunch of status info
	// the UDRE0 bit of the register is the USART Data Register Empty
	// bit, and will be high when the register is empty and ready to recieve new data to send.
	while (!(peekR(UCSR0A, RXC0)))  //while not ready to transieve
	{
		//no op;
	}

	return UDR0; // once recieved return the char
}
















#endif